%---------------------------
%discrete event simulation
%the time is counted in microseconds
%1 second=1000  milliseconds=10^6 microseconds
%6-node ring network
%EBSL: kick the tail segments of the burst with lowest priority
%---------------------------
clear all;
format long;
propagation_speed=3*1e8/1e6;
%propagation_speed=1e20;
switching_delay=0; %usually smaller than 1, so ignore
offset_time=10; %micorseconds
packet_length=1250*8; %bits
%packet_length=1250;
%average burst length=2.5MB
packet_no_ave=250;
transmission_speed=10*1e9/1e6; %bit/micros

Simulation=2e6;
node=6;
L=6;
OD_no=6;
C=50;
%OD_path=zeros(30,7);
OD_path=[ 1 2 3 0 0 0 0;
     2 3 4 0 0 0 0; 
     3 4 5 0 0 0 0; 
    4 5 6 0 0 0 0; 
     5 6 1 0 0 0 0; 
    6 1 2 0 0 0 0];
link_length=[2*1e5  2*1e5  2*1e5  2*1e5  2*1e5  2*1e5]*5; %the length of each link

Lambda=[0.045 0.05 0.06 0.07 0.08 0.1 0.12 0.14 0.2 0.25 0.3];  % offered load to each source-destination node
 

for traf=1:length(Lambda)
    total_block=zeros(1,traf);
    lambda=ones(1,OD_no)*Lambda(traf);  %busrts/micosecond
    
    %propagation_daley_link=link_length./propagation_speed;
    buffer=0;
    link_use=zeros(L,C,1+buffer,8);%layer-1 1-occupy or not; 2-OD_pair; 3- hop no; 4-left no. of hop time; 5-finish time 6-reservation no 7-burst length (no. of packets) 8-no of left bits ;
    %layer-2 1-waiting or not;
    %2-OD_pair; 3-hop no.; 4 -min start time;
    %5-max leave time 6-reservation_no 7-burst length (no. of packets)  8-no of left bits ;
    buffer2=40000;
    reservation_queque=zeros(OD_no,buffer2,6);%1-occupy or not 2-next hop no 3 next hop arrival time 4-reservation_no 5-burst length (no. of packets) 6-no of left bits
    
    block=zeros(1,OD_no);
    departure=zeros(1,OD_no);
    traffic=lambda*(packet_length*packet_no_ave/transmission_speed);
    link_block=zeros(1,L);
    link_departure=zeros(1,L);
    good_calls=zeros(Simulation,9); %1-hop no 2- ariival time 3- deparutre time; 4,7==1; 5,8=2; 6,9==3;
    
    
    
    arrivelIntervalTime_control(1:Simulation) = exprnd(1/(sum(lambda)),[1,Simulation]);  % ariival of the control packets, follow possion distribution
    nextNewArrivalTime=arrivelIntervalTime_control(1)+offset_time;
    rd=rand();
    part_arrival=lambda(1);
    finish=0;
    count=1;
    while (finish==0)
        if rd<part_arrival/sum(lambda)
            OD_pair=count;
            finish=1;
        else
            count=count+1;
            part_arrival=part_arrival+lambda(count);
        end
    end
    packet_no=floor(exprnd(packet_no_ave,1));
    path_length=length(find(OD_path(OD_pair,:)>0));  %find the length of the path;
    % calculate the happen time of the next event of different kinds
    nextArrivalTime=nextNewArrivalTime+offset_time+link_length(OD_path(OD_pair,1))/propagation_speed; % time of the burst arrival the next hop
    nextBurst_come=nextArrivalTime;  %the time of the next burst arriving to a hop (not a new arrival)
    nextBurstEvent_finish=nextArrivalTime+packet_no*packet_length/transmission_speed;  %the time of the burst finish serving in this hop
    %record in the table
    reservation_queque(OD_pair,1,:)=[1 2 nextArrivalTime 1 packet_no packet_no*packet_length];
    link_use(OD_path(OD_pair,1),1,1,:)=[1 OD_pair 1 2 nextBurstEvent_finish 1 packet_no packet_no*packet_length];
    reservation_count=1;
    good_calls(reservation_count,1)=OD_path(OD_pair,1);
    good_calls(reservation_count,2)=nextNewArrivalTime;
    state_probability=zeros(L,C+1);
    %segBurst_com=1e20;
    
    
    %four kinds of event (1) the first reservation of a new control packet (2) a new reservation finish (not in the first hop) (3) a burst comes into a %%hop (4) a burst finish serving in a hop
    
    arrival=1;
    nextNewArrivalTime=nextNewArrivalTime+arrivelIntervalTime_control(arrival+1);
    while arrival<Simulation-1
        
        if nextNewArrivalTime<=nextBurst_come && nextNewArrivalTime<=nextBurstEvent_finish
            %a new control packet finish reserve the resource ( reserving resources ), a new call comes into the network
            %generate the OD no, path length of the new control packet and the arrival time of the new burst
            
            for iii=1:L
                x=sum(link_use(iii,:,1,1));
                state_probability(iii,x+1)=state_probability(iii,x+1)+1;
            end
            
            rd=rand();
            part_arrival=lambda(1);
            finish=0;
            count=1;
            while (finish==0)
                if rd<part_arrival/sum(lambda)
                    OD_pair=count;
                    finish=1;
                else
                    count=count+1;
                    part_arrival=part_arrival+lambda(count);
                end
            end
            packet_no=floor(exprnd(packet_no_ave,1)); %generate the number of packets of the burst
            burst_finish_time=nextNewArrivalTime+packet_no*packet_length/transmission_speed;
            nextHopArrivalTime=nextNewArrivalTime+offset_time+link_length(OD_path(OD_pair,1))/propagation_speed;
            path_length=length(find(OD_path(OD_pair,:)>0));
            
            %find free path
            occupy=link_use(OD_path(OD_pair,1),:,1,1);
            isFree=find(occupy(:)==0);
            if isempty(isFree)
                %find if can kick
                xx=find(link_use(OD_path(OD_pair,1),:,1,4)>=path_length);
                
                if isempty(xx)
                    %no free slots, block the reservation
                    block(OD_pair)=block(OD_pair)+packet_no;
                    link_block(OD_path(OD_pair,1))=link_block(OD_path(OD_pair,1))+packet_no;
                else
                    %the call finds a channel with lower priority call serves,
                    %so the new call kicks the old one
                    max_length=max(link_use(OD_path(OD_pair,1),xx,1,4));
                    xxx=find(link_use(OD_path(OD_pair,1),:,1,4)==max_length);
                    
                    channel=xxx(floor(rand()*length(xxx)+1));
                    template=link_use(OD_path(OD_pair,1),channel,1,:);
                    %fit the new call
                    reservation_count=reservation_count+1;
                    link_use(OD_path(OD_pair,1),channel,1,:)=[1 OD_pair 1 2 burst_finish_time  reservation_count packet_no packet_no*packet_length];
                    %update reservation_queque
                    slot=min(find(reservation_queque(OD_pair,:,1)==0));
                    %update good_calls
                    good_calls(reservation_count,1)=OD_path(OD_pair,1);
                    good_calls(reservation_count,2)=nextNewArrivalTime;
                    reservation_queque(OD_pair,slot,:)=[1 2  nextHopArrivalTime reservation_count packet_no packet_no*packet_length];
                    % update the next event time of different kinds
                    if nextBurst_come>nextHopArrivalTime
                        nextBurst_come=nextHopArrivalTime;
                    end
                    
                    if nextBurstEvent_finish>burst_finish_time
                        nextBurstEvent_finish=burst_finish_time;
                    end
                    % deal with the segmented burst
                    %the rest of the burst will drop and the reservation_queque
                    %and the link_use need to update
                    seg_reservation_no=template(6);
                    seg_reservation=find(reservation_queque(:,:,4)==seg_reservation_no);
                    slot=floor((seg_reservation-1)./OD_no)+1;
                    seg_OD=seg_reservation-(slot-1)*OD_no;
                    seg_finish_time=template(5);
                    time=seg_finish_time-nextNewArrivalTime;
                    bits_left=template(8)-floor(time*transmission_speed)-1;
                    link_block(OD_path(OD_pair,1))=link_block(OD_path(OD_pair,1))+floor(time*transmission_speed/packet_length)+1;
                    link_departure(OD_path(OD_pair,1))=link_departure(OD_path(OD_pair,1))+floor(bits_left/packet_length);
                    if (reservation_queque(seg_OD,slot,6)>bits_left)
                        reservation_queque(seg_OD,slot,6)=bits_left;
                    end
                    
                    %need to update the link_use
                    calls=find(link_use(:,:,:,6)==seg_reservation_no);
                    seg_place=1; % always 1
                    seg_channel=floor((calls-1)/L+1);
                    seg_L=calls-(seg_channel-1)*L;
                    %the bits on the hop before this hop just let it finish, do not need to update,
                    for i=1:length(calls)
                        if link_use(seg_L(i),seg_channel(i),seg_place(i),3)>template(3) && link_use(seg_L(i),seg_channel(i),seg_place(i),8)>bits_left
                            %update the information of the hops after this seg hop
                            time_save=(link_use(seg_L(i),seg_channel(i),seg_place(i),8)-bits_left)/transmission_speed;
                            link_use(seg_L(i),seg_channel(i),seg_place(i),8)=bits_left;
                            link_use(seg_L(i),seg_channel(i),seg_place(i),5)=link_use(seg_L(i),seg_channel(i),seg_place(i),5)-time_save;
                            if nextBurstEvent_finish>link_use(seg_L(i),seg_channel(i),1,5)
                                    nextBurstEvent_finish=link_use(seg_L(i),seg_channel(i),1,5);
                                end
                        end
                    end
                    
                end
                
            else
                %free time slots, random choose one to fit
                reservation_count=reservation_count+1;
                good_calls(reservation_count,1)=OD_path(OD_pair,1);
                good_calls(reservation_count,2)=nextNewArrivalTime;
                channel=isFree(floor(rand()*length(isFree)+1));  %random select a free channel
                link_use(OD_path(OD_pair,1),channel,1,:)=[1 OD_pair 1 2 burst_finish_time  reservation_count packet_no packet_no*packet_length];
                %update reservation_queque
                slot=min(find(reservation_queque(OD_pair,:,1)==0));
                reservation_queque(OD_pair,slot,:)=[1 2  nextHopArrivalTime reservation_count packet_no packet_no*packet_length];
                % update the next event time of different kinds
                if nextBurst_come>nextHopArrivalTime
                    nextBurst_come=nextHopArrivalTime;
                end
                
                if nextBurstEvent_finish>burst_finish_time
                    nextBurstEvent_finish=burst_finish_time;
                end
            end
            arrival=arrival+1;
            nextNewArrivalTime=nextNewArrivalTime+arrivelIntervalTime_control(arrival+1);
        elseif (nextBurst_come<=nextBurstEvent_finish && nextBurst_come<=nextNewArrivalTime )
            % a control packet finish reserve in next hop
            %find the reservation number
            reservation=find(reservation_queque(:,:,3)==nextBurst_come);
            for k=1:length(reservation)
                %find the information of this reservation
                slot=floor((reservation(k)-1)/OD_no)+1;
                OD_pair=reservation(k)-(slot-1)*OD_no;
                hop_no=reservation_queque(OD_pair,slot,2);
                thisHop=OD_path(OD_pair,hop_no);
                path_length=length(find(OD_path(OD_pair,:)>0));
                burstArrivalTime=nextBurst_come;
                reservation_no=reservation_queque(OD_pair,slot,4);
                packet_no=reservation_queque(OD_pair,slot,5);
                left_burst_length=reservation_queque(OD_pair,slot,6);
                %calculate the information of the reservation
                burstArrivalNextHopTime=burstArrivalTime+link_length(thisHop)/propagation_speed+offset_time;
                burst_finish_time=burstArrivalTime+left_burst_length/transmission_speed;
                %find if the next hop has free slots
                occupy=link_use(thisHop,:,1,1);
                isFree=find(occupy(:)==0);
                if isempty(isFree)
                    xx=find(link_use(thisHop,:,1,4)>=3-hop_no+1); %find if there is lower priority calls
                    if isempty(xx)
                        %the reservation is not success
                        block(OD_pair)=block(OD_pair)+packet_no;
                        link_block(thisHop)=link_block(thisHop)+floor(left_burst_length/packet_length);
                        reservation_queque(OD_pair,slot,:)=[0 0 0 0 0 0];
                        good_calls(reservation_no,:)=[0 0 0 0 0 0 0 0 0];
                    else
                        %randomly kick one lower priority call
                        max_length=max(link_use(thisHop,xx,1,4));
                        xxx=find(link_use(thisHop,:,1,4)==max_length);
                    
                        channel=xxx(floor(rand()*length(xxx)+1));
                        template=link_use(thisHop,channel,1,:);
                        %fit the call
                        link_use(thisHop,channel,1,:)=[1 OD_pair hop_no path_length-hop_no burst_finish_time  reservation_no packet_no left_burst_length];
                        %update good_calls
                        good_calls(reservation_no,hop_no*3-2)=thisHop;
                        good_calls(reservation_no,hop_no*3-1)=burstArrivalTime;
                        %update the reservation_queque
                        if hop_no==path_length
                            %the control packet finish reservation
                            reservation_queque(OD_pair,slot,:)=[0 0 0 0 0 0];
                        else
                            reservation_queque(OD_pair,slot,:)=[1 hop_no+1 burstArrivalNextHopTime reservation_no packet_no left_burst_length];
                        end
                        %update the time of events
                        if nextBurstEvent_finish>burst_finish_time
                            nextBurstEvent_finish=burst_finish_time;
                        end
                        
                        % deal with the segmented burst
                        %the rest of the burst will drop and the reservation_queque
                        %and the link_use need to update
                        seg_reservation_no=template(6);
                        seg_reservation=find(reservation_queque(:,:,4)==seg_reservation_no);
                        slot=floor((seg_reservation-1)./OD_no)+1;
                        seg_OD=seg_reservation-(slot-1)*OD_no;
                        seg_finish_time=template(5);
                        time=seg_finish_time-nextBurst_come;
                        bits_left=template(8)-floor(time*transmission_speed)-1;
                        link_block(thisHop)=link_block(thisHop)+floor(time*transmission_speed/packet_length)+1;
                        link_departure(thisHop)=link_departure(thisHop)+floor(bits_left/packet_length);
                        if (reservation_queque(seg_OD,slot,6)>bits_left)
                            reservation_queque(seg_OD,slot,6)=bits_left;
                        end
                        %need to update the link_use
                        calls=find(link_use(:,:,:,6)==seg_reservation_no);
                        seg_place=1; % always 1
                        seg_channel=floor((calls-1)/L+1);
                        seg_L=calls-(seg_channel-1)*L;
                        %the bits on the hop before this hop just let it finish, do not need to update,
                        for i=1:length(calls)
                            if link_use(seg_L(i),seg_channel(i),seg_place(i),3)>template(3) && link_use(seg_L(i),seg_channel(i),seg_place(i),8)>bits_left
                                %update the information of the hops after this seg hop
                                time_save=(link_use(seg_L(i),seg_channel(i),seg_place(i),8)-bits_left)/transmission_speed;
                                link_use(seg_L(i),seg_channel(i),seg_place(i),8)=bits_left;
                                link_use(seg_L(i),seg_channel(i),seg_place(i),5)=link_use(seg_L(i),seg_channel(i),seg_place(i),5)-time_save;
                                if nextBurstEvent_finish>link_use(seg_L(i),seg_channel(i),1,5)
                                    nextBurstEvent_finish=link_use(seg_L(i),seg_channel(i),1,5);
                                end
                            end
                        end
                    end
                    
                    
                else
                    %the reservation is sucess
                    channel=isFree(floor(rand()*length(isFree)+1));
                    link_use(thisHop,channel,1,:)=[1 OD_pair hop_no path_length-hop_no burst_finish_time  reservation_no packet_no left_burst_length];
                    good_calls(reservation_no,hop_no*3-2)=thisHop;
                    good_calls(reservation_no,hop_no*3-1)=burstArrivalTime;
                    %update the reservation_queque
                    if hop_no==path_length
                        %the control packet finish reservation
                        reservation_queque(OD_pair,slot,:)=[0 0 0 0 0 0];
                    else
                        reservation_queque(OD_pair,slot,:)=[1 hop_no+1 burstArrivalNextHopTime reservation_no packet_no left_burst_length];
                    end
                    %update the time of events
                    if nextBurstEvent_finish>burst_finish_time
                        nextBurstEvent_finish=burst_finish_time;
                    end
                end
            end
            control_event=find(reservation_queque(:,:,3)>0);
            if isempty(control_event)
                nextBurst_come=1e40;
            else
                slot1=floor((control_event-1)./OD_no)+1;
                OD_pair=control_event-(slot1-1)*OD_no;
                nextBurst_come=reservation_queque(OD_pair(1),slot1(1),3);
                for i=2:length(slot1)
                    if nextBurst_come>reservation_queque(OD_pair(i),slot1(i),3)
                        nextBurst_come=reservation_queque(OD_pair(i),slot1(i),3);
                    end
                end
            end
            
            
        else
            %a burst finish serving in a hop and release the resource
            burst=find(link_use(:,:,1,5)==nextBurstEvent_finish);
            for k=1:length(burst)
                channel=floor((burst(k)-1)/L)+1;
                link=burst(k)-(channel-1)*L;
                OD_pair=link_use(link,channel,1,2);
                %check if this hop is the last hop of the path (if so, the call leave the network
                hop_no=link_use(link,channel,1,3);
                packet_no=link_use(link,channel,1,7);
                left_burst_length=link_use(link,channel,1,8);
                path_length=length(find(OD_path(OD_pair,:)>0));
                link_departure(link)=link_departure(link)+floor(left_burst_length/packet_length);
                if (hop_no==path_length)
                    received_packet=floor(left_burst_length/packet_length);
                    departure(OD_pair)=departure(OD_pair)+received_packet;
                    block(OD_pair)=block(OD_pair)+(packet_no-received_packet);
                    %update the finish time of the good_calls
                    reservation_no=link_use(link,channel,1,6);
                    if received_packet>0
                    efficient_service_time=received_packet*packet_length/transmission_speed;
                    good_calls(reservation_no,3)=good_calls(reservation_no,2)+efficient_service_time;
                    good_calls(reservation_no,6)=good_calls(reservation_no,5)+efficient_service_time;
                    good_calls(reservation_no,9)=good_calls(reservation_no,8)+efficient_service_time;
                    else
                        good_calls(reservation_no,:)=[0 0 0 0 0 0 0 0 0];
                    end
                end
                %release the resource
                link_use(link,channel,1,:)=[0 0 0 0 0 0 0 0];
            end
            % update the events time
            burst_event=find(link_use(:,:,1,5)>0);
            if isempty(burst_event)
                nextBurstEvent_finish=1e20;
            else
                channel=floor((burst_event-1)./L)+1;
                link=burst_event-(channel-1)*L;
                nextBurstEvent_finish=link_use(link(1),channel(1),1,5);
                for i=2:length(channel)
                    if nextBurstEvent_finish>link_use(link(i),channel(i),1,5)
                        nextBurstEvent_finish=link_use(link(i),channel(i),1,5) ;
                    end
                end
            end
        end
    end
    total_block(traf)=sum(block./(block+departure))/6
    
    
    
    utilization=zeros(L,1);
    for i=1:L
        state_sum=sum(state_probability(i,:));
        state_probability(i,:)=state_probability(i,:)/state_sum;
        average=0;
        for j=1:C+1
            average=average+state_probability(i,j)*(j-1)/C;
        end
        utilization(i)=utilization(i)+average;
    end
    
    U(traf)=sum(utilization)/L
    link_b=link_block./(link_block+link_departure);
    
    
    
    
    goodU_link=zeros(1,L);
    
    carried_link=zeros(1,L);
    arrival_link=zeros(1,L);
    offered_link=zeros(1,L);
    for i=1:OD_no
        for j=1:3
            goodU_link(OD_path(i,j))=goodU_link(OD_path(i,j))+(1-block(i)/(block(i)+departure(i)))*traffic(i);
        end
    end
    for i=1:OD_no
        offered_link(OD_path(i,1))=offered_link(OD_path(i,1))+traffic(i)*(1-link_b(OD_path(i,1)));
        offered_link(OD_path(i,2))=offered_link(OD_path(i,2))+traffic(i)*(1-link_b(OD_path(i,1)))*(1-link_b(OD_path(i,2)));
        offered_link(OD_path(i,3))=offered_link(OD_path(i,3))+traffic(i)*(1-link_b(OD_path(i,1)))*(1-link_b(OD_path(i,2)))*(1-link_b(OD_path(i,3)));
    end
    U_new(traf)=sum(offered_link./C)/L
    goodutilization=goodU_link/C;
    goodU(traf)=sum(goodutilization)/L
    badU(traf)=U(traf)-goodU(traf)
    goodput(traf)=sum((1-block./(block+departure)).*traffic(i))
    
%     for i=Simulation:-1:1
%         if good_calls(i,1)==0
%             good_calls(i,:)=[];
%         end
%     end
    no_good_calls=length(good_calls(:,1));
    inspector_interval=exprnd(1.2/(sum(lambda)),[1,Simulation/2]);
    inspector=zeros(1,Simulation/2);
    inspector(1)=1/(sum(lambda))*10000;
    for i=2:Simulation/2
        inspector(i)=inspector(i-1)+inspector_interval(i-1);
    end
    good_utilization_count=zeros(Simulation/2,L);
    
    for i=1:no_good_calls
        if good_calls(i,9)==0
        else
            x=min(find(inspector(:)>=good_calls(i,2)));
            y=max(find(inspector(:)<good_calls(i,3)));
            if x<=y
                for j=x:y
                    good_utilization_count(j,good_calls(i,1))=good_utilization_count(j,good_calls(i,1))+1;
                end
            end
            x=min(find(inspector(:)>=good_calls(i,5)));
            y=max(find(inspector(:)<good_calls(i,6)));
            if x<=y
                for j=x:y
                    good_utilization_count(j,good_calls(i,4))=good_utilization_count(j,good_calls(i,4))+1;
                end
            end
            x=min(find(inspector(:)>=good_calls(i,8)));
            y=max(find(inspector(:)<good_calls(i,9)));
            if x<=y
                for j=x:y
                    good_utilization_count(j,good_calls(i,7))=good_utilization_count(j,good_calls(i,7))+1;
                end
            end
        end
    end
    
    efficient_state_probability=zeros(L,C+1);
    for i=1:L
        for j=1:Simulation/2
            state=good_utilization_count(j,i);
            efficient_state_probability(i,state+1)=efficient_state_probability(i,state+1)+1;
        end
    end
    efficient_utilization=zeros(1,L);
    for i=1:L
        state_sum=sum(efficient_state_probability(i,:));
        efficient_state_probability(i,:)=efficient_state_probability(i,:)/state_sum;
        average=0;
        for j=1:C+1
            average=average+efficient_state_probability(i,j)*(j-1)/C;
        end
        efficient_utilization(i)=efficient_utilization(i)+average; 
    end
    EU(traf)=sum(efficient_utilization)/L
    
       fid = fopen('EBSL.txt','wt'); 
    fprintf(fid,'%c','total_block=[ ');
    for k=1:traf 
         p_B=num2str(total_block(k));
        fprintf(fid,'%c',p_B);  
        fprintf(fid,'%c',' '); 
    end
    fprintf(fid,'%c',' ];');
     fprintf(fid,'%c\n',' '); 
    fprintf(fid,'%c','U=[ ');
    for k=1:traf;
         p_U=num2str(U(k));
        fprintf(fid,'%c',p_U);  
        fprintf(fid,'%c',' '); 
    end
    fprintf(fid,'%c',' ];');
    fprintf(fid,'%c\n',' '); 
    fprintf(fid,'%c','goodput=[ ');
    for k=1:traf 
         p_g=num2str(goodput(k));
        fprintf(fid,'%c',p_g);  
        fprintf(fid,'%c',' '); 
    end
    fprintf(fid,'%c',']; ');
    fprintf(fid,'%c\n',' '); 
    fprintf(fid,'%c','EU=[ ');
    for k=1:traf
         p_EU=num2str(EU(k));
        fprintf(fid,'%c',p_EU);  
        fprintf(fid,'%c',' '); 
    end
    fprintf(fid,'%c',']; ');
     fprintf(fid,'%c\n',' ');
     fprintf(fid,'%c\n',' ');
     fclose(fid);
    
    
end


